/* $Workfile: Dsp55Demo2.c $   $Revision: 18 $    $Modtime: 3/02/07 6:30p $  */
/*--------------------------------------------------------------------------*\

  Odin TeleSystems Inc.

  SAA-1012-1 OTX DSP C55x Software Development Kit (SDK)

  Demo2

  This demo illustrates how to send and receive serial user data. It receives
  u-law encoded voice samples. It loops these samples back in the transmit
  direction after performing a gain adjustment. The gain adjustment value
  is selected by passing an IO Control code to this application.

  Use together with Host application Dsp55Demo2.c

\*--------------------------------------------------------------------------*/

#include <string.h>   // for memset()     

#include <Dsp55Apicfg.h> // automatically generated by the gconfgen tool
#include <csl.h>
#include <csl_dma.h>
#include <csl_mcbsp.h>
#include <csl_irq.h>
#include <csl_chip.h>
#include <csl_timer.h>

#include <Dsp55Api.h>
#include <Dsp55Ioctl.h>


#include "Dsp55Demo2Io.h"  


/*--------------------------------------------------------------------------*\
  Macros
\*--------------------------------------------------------------------------*/
#define DEMO_MAX_TIMESLOTS_PER_FRAME 128

unsigned short g_nFrameCounter = 0;

/*--------------------------------------------------------------------------*\
  Private data
\*--------------------------------------------------------------------------*/

static unsigned short g_sqFrameSample[DEMO_MAX_TIMESLOTS_PER_FRAME];  // Temporary storage area for one frame

unsigned short g_hHandle = OTXDSP_INVALID_HANDLE;  // Handle passed to us from the host user application

static short g_law = OTXDSP_ULAW;

static short g_nTxGain = 0;

static short g_nRxTimeSlot = 1;
static short g_nTxTimeSlot = 0; 

#if 0
static short g_bRxEnabled = OTXDSP_FALSE;
static short g_bTxEnabled = OTXDSP_FALSE;
#else
static short g_bRxEnabled = OTXDSP_TRUE;
static short g_bTxEnabled = OTXDSP_TRUE;
#endif

static short g_sqRxData[OTXDSP_SDK_FRAMES_PER_BUF];
static short g_sqTxData[OTXDSP_SDK_FRAMES_PER_BUF];
static short g_iTxData = 0;
static short g_iRxData = 0;

static short g_bStreamRxData = OTXDSP_FALSE;

static short g_bLoopAllTs = OTXDSP_FALSE;

Demo2NotifyGetDebugDataS g_debugDataS;

static Uint32 g_nDmaRxFrameCounter = 0;
static Uint32 g_nDmaTxFrameCounter = 0;

static volatile long  g_blinkCount      = 0;
static volatile long  g_blinkCountMax   = 50000L; //100000L;
static volatile short g_blinkXfVal      = OTXDSP_FALSE;

// For testing the external memory
#ifdef EXTMEM_AVAILABLE
#pragma DATA_SECTION(g_sqBigBuf,"extMem")
unsigned short g_sqBigBuf[3000]; 
#endif

#if 1 // mainly for debugging
static unsigned short g_iTemp = 0;
static short g_bRxDataAvailable = OTXDSP_FALSE;
#endif

#if 1 // Temporary
extern unsigned short g_debugData[8];
extern unsigned short g_debugBuf[32];
#endif


/*--------------------------------------------------------------------------*\
  Function Prototypes
\*--------------------------------------------------------------------------*/
static void ProcessRxSamples(void);
static void GenerateTxSamples(void);
static void UnpackFrame(unsigned short *pFrame);
static void PackFrame(unsigned short *pFrame);
#ifdef EXTMEM_AVAILABLE
static unsigned short DemoCheckExternalMemory(
  Uint32 *nDataErrors,
  Uint32 *nMemErrors
);
#endif


/*--------------------------------------------------------------------------*\
  @func
  
    main() is called from _c_int00 which is the entry point for the program.
    _c_int00 will first initialize all global data before running main().
    
    The function CSL_cfgInit() will also be called before main() is called.
    
    This function will initialize global data and peripherals and then terminate
    to allow the DSP/BIOS tasks to run.
    
\*--------------------------------------------------------------------------*/
main() {    
  //unsigned short readVal; 
  
  LOG_printf(&trace, "Demo program started\n");  
  
  memset(g_sqRxData, 0, OTXDSP_SDK_FRAMES_PER_BUF);
  memset(g_sqTxData, 0, OTXDSP_SDK_FRAMES_PER_BUF);

  // Run Standard OTX DSP SDK API Initializations
  OtxDspInit();

  // Initialize application specific data
  memset(&g_debugDataS, 0, sizeof(Demo2NotifyGetDebugDataS));  

  LOG_printf(&trace, "main() terminating... DSP BIOS tasks should be running.\n");  

}  

  
/*--------------------------------------------------------------------------*\
  @func

    This function is the main process task thread. It will go to sleep
    until a semaphore is posted from an ISR, which will wake up the thread
    to allow it to execute and process data (highway receive data, 
    highway transmit data, and control data from the host).
    
\*--------------------------------------------------------------------------*/
Void TskDataProcessing(Arg value_arg)
{
  Int value = ArgToInt(value_arg);
  Bool semRc = 1; 

  LOG_printf(&trace, "Data Processing task started (Arg=%d)\n", value);  

  // Loop forever
  for(;;) {   
                                          
    // check if we have RX samples available, process them. E.g. run detection algorithms on them
    if (g_bGotDmaRxInterrupt) {  // g_bGotDmaRxInterrupt updated by the interrupt service routine
      g_bGotDmaRxInterrupt = OTXDSP_FALSE;
      OtxDspServiceRxSerialPort();
    }

    // check if more samples are needed to be transmitted, generate more samples.
    if (g_bGotDmaTxInterrupt) {  // g_bGotDmaTxInterrupt updated by the interrupt service routine
      g_bGotDmaTxInterrupt = OTXDSP_FALSE;
      OtxDspServiceTxSerialPort();
    }

    // check if the host has requested our attention. If so, service his request
    if (g_bGotHostPortInterrupt) {  // g_bGotHostPortInterrupt updated by the interrupt service routine
      g_bGotHostPortInterrupt = OTXDSP_FALSE;
    
      if (*g_h2dCmdAddrP != OTXDSP_SDK_IO_NO_COMMAND) { 
        // We have an IO Control command to process
        OtxDspServiceIoControl();
      }
    }


    // Sleep until we get a hardware interrupt (serial port or host port)
    semRc = SEM_pend(&semProcessing, 20); //SYS_FOREVER); 
    
    if (!semRc) {
      g_debugDataS.nData1++;
    }


  }  /* for */

}


/*--------------------------------------------------------------------------*\
  @func
    
    Idle task. 
  
    Blink the heart-beat LED  (green LED2).

\*--------------------------------------------------------------------------*/
Void IdlBlink(Arg value_arg)
{
  Int value = ArgToInt(value_arg); 

  LOG_printf(&trace, "Blink LED task started (Arg=%d)\n", value);  

  if (g_blinkCount++ >= g_blinkCountMax) {
    g_blinkCount = 0;
  
    if (g_blinkXfVal) {
      g_blinkXfVal = OTXDSP_FALSE;
      C55X_SET_LED2_ON;   
    } else {
      g_blinkXfVal = OTXDSP_TRUE;
      C55X_SET_LED2_OFF;   
    } 
        
  }
}



/*--------------------------------------------------------------------------*\
  @func

    Call-back routine called by the Interrupt Service Routine when the
    host has issued a command to the HPI and requests our attention.

    This function is called from the processing task.
    TO BE RE-IMPLEMENTED BY THE USER.
\*--------------------------------------------------------------------------*/
void OtxDspServiceIoControl(
  void
)
{
  //unsigned short nDataLen      = *((unsigned short *) (OTXDSP_SDK_IOCTL_DATA_LEN_ADDR/2));
  unsigned short hHandle       = *((unsigned short *) (OTXDSP_SDK_IOCTL_HANDLE_ADDR/2));
  unsigned short nTaskRef      = *((unsigned short *) (OTXDSP_SDK_IOCTL_TASKREF_ADDR/2));
  //short          bDataComplete = *((unsigned short *) (OTXDSP_SDK_IOCTL_CODE_ADDR/2)) & OTXDSP_SDK_MASK_IO_DATA_COMPLETE;
  //short          bDataPadded   = *((unsigned short *) (OTXDSP_SDK_IOCTL_CODE_ADDR/2)) & OTXDSP_SDK_MASK_IO_DATA_PADDED;
  unsigned short nCode         = *((unsigned short *) (OTXDSP_SDK_IOCTL_CODE_ADDR/2)) & OTXDSP_SDK_MASK_IO_CODE;


  // Standard SDK IO Control codes
  if (nCode == OTXDSP_SDK_IO_CREATE) {
    if (g_hHandle == OTXDSP_INVALID_HANDLE) {
      // The handle is not set
      g_hHandle = hHandle;
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // This demo only supports one instance of the logical device. Return an error code.
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_OUT_OF_RESOURCES);
    }

  } else if (nCode == OTXDSP_SDK_IO_CLOSE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_ENABLE) {
    g_bRxEnabled = OTXDSP_TRUE;
    g_bTxEnabled = OTXDSP_TRUE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_DISABLE) {
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_CONNECT) {
    OtxDspSdkIoInConnectS *pInS = (OtxDspSdkIoInConnectS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2);
    short i;
    short bEvenTimeSlot;
    short nWordSlot;       

    // Save these values in the debug struct. To be retrieved when the DSPDEMO1_IOCTL_GET_DEBUG_DATA control is issued
    //g_debugDataS.nData0     = pInS->nDirection;

    if (pInS->nCapacity == 64) {
      // Only 64kbit/s connections are supported by this demo program
      for (i = 0; i < OTXDSP_SDK_WORDSLOTS_PER_HIGHWAY; i++) {
        if (pInS->sqConnectMask[i]) {
          break;
        }
      }

      if (pInS->nDirection == OTXDSP_SDK_RX_DIRECTION) {
        // Connection data for the Receiver
        nWordSlot = i;
        // Even or odd timeslot
        bEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;
        // 
        g_nRxTimeSlot = nWordSlot*2 + (bEvenTimeSlot ? 0 : 1);

        //g_debugDataS.nData1 = nWordSlot;
        //g_debugDataS.nData2 = bEvenTimeSlot;

      } else {
        // Connection data for the Transmitter
        nWordSlot = i;
        // Even or odd timeslot
        bEvenTimeSlot = (pInS->sqConnectMask[i] == 0x00FF) ? OTXDSP_TRUE : OTXDSP_FALSE;
        // 
        g_nTxTimeSlot = nWordSlot*2 + (bEvenTimeSlot ? 0 : 1);
      }

#if 1 // Temporary since we dont get ENABLE yet
      g_bRxEnabled = OTXDSP_TRUE;
      g_bTxEnabled = OTXDSP_TRUE;
#endif

      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);
    } else {
      // Wrong Pipe capacity. Only 64kbit/s connections are supported by this demo program
      OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
    }


  } else if (nCode == OTXDSP_SDK_IO_DISCONNECT) {
    // No specific action. Just disable.
    g_bRxEnabled = OTXDSP_FALSE;
    g_bTxEnabled = OTXDSP_FALSE;
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_RESET) {
    // No special reset action in this example. Just clear the debug data struct
    g_debugDataS.nData0 = 0;
    g_debugDataS.nData1 = 0;
    g_debugDataS.nData2 = 0;
    g_debugDataS.nData3 = 0;
    g_debugDataS.nData4 = 0;
    g_debugDataS.nData5 = 0;
    g_debugDataS.nData6 = 0;
    g_debugDataS.nData7 = 0;

    g_iTemp = 0;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_INIT_BSP) {
    //OtxDspInitSerialPort(); 
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == OTXDSP_SDK_IO_START_PROGRAM) {
#if 0 // Old way is still supported by the driver
    unsigned short dspRc;
    dspRc = OtxDspStartProgram((OtxDspSdkIoInStartProgramS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2));
    OtxDspCommandCompleteRc(hHandle, dspRc);
#else // Provide more information back to the host driver upon startup
    OtxDspSdkIoOutStartProgramS sOut;

    sOut.nResult = OtxDspStartProgram((OtxDspSdkIoInStartProgramS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2));

    sOut.nEventBufAddr     = OTXDSP_SDK_EVENT_BUF_ADDR;
    sOut.nEventBufLen      = OTXDSP_SDK_EVENT_BUF_LEN;
    sOut.nIoctlCodeAddr    = OTXDSP_SDK_IOCTL_CODE_ADDR;
    sOut.nIoctlDataMaxSize = 0; //(unsigned short)(OTXDSP_SDK_IOCTL_DATA_MAX_SIZE);

    sOut.nReserved0 = sOut.nReserved1 = sOut.nReserved2 = sOut.nReserved3 = sOut.nReserved4 = 0;

    if (sOut.nResult == OTXDSP_SDK_S_OK) {
      C55X_SET_LED0_ON; // Turn on LED0 (Green AUX LED on Sleipnir)
    } else {
      C55X_SET_LED1_ON; // Turn on LED1 (Red AUX LED on Sleipnir)
    }

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(OtxDspSdkIoOutStartProgramS));
#endif

    
  } else if (nCode == OTXDSP_SDK_IO_INIT_TIMER) {
    unsigned short dspRc;
    dspRc = OtxDspInitTimer();
    OtxDspCommandCompleteRc(hHandle, dspRc);

  } else if (nCode == OTXDSP_SDK_IO_GET_FRAME_BUFFER_COUNTER) {
    OtxDspSdkIoOutGetFrameBufferCounter sOut;
    sOut.nResult = OTXDSP_SDK_S_OK;
    sOut.nFrameBufferCounter = g_nFrameBufferCounter;
    sOut.nReserved0          = g_nFrameCounter;
    sOut.nReserved1          = 0;
    sOut.nReserved2          = 0;
    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(OtxDspSdkIoOutGetFrameBufferCounter));
    

  // Application specific IO Control Codes
  } else if (nCode == DEMO2_IOCTL_SET_TX_GAIN) {
    Demo2IoInSetTxGainS *pInS = (Demo2IoInSetTxGainS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2);

    g_nTxGain = pInS->nTxGain;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_STREAM_RX_DATA) {
    Demo2IoInStreamRxDataS *pInS = (Demo2IoInStreamRxDataS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2);

    g_bStreamRxData = pInS->bStreamRxData;

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_CONTROL_LED) {
    Demo2IoInControlLedS *pInS = (Demo2IoInControlLedS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2);
    unsigned short dspRc = OTXDSP_SDK_S_OK;
    
    switch (pInS->nLed) {
    case 0:
      if (pInS->bOn) {
        C55X_SET_LED0_ON;
      } else {
        C55X_SET_LED0_OFF;
      }
      break;
    case 1:
      if (pInS->bOn) {
        C55X_SET_LED1_ON;
      } else {
        C55X_SET_LED1_OFF;
      }
      break;
    default:
      dspRc = OTXDSP_SDK_E_INVALIDARG;
      break;
    }

    OtxDspCommandCompleteRc(hHandle, dspRc);

  } else if (nCode == DEMO2_IOCTL_GET_SAMPLE) {
    Demo2IoOutGetSampleS sOut;

    sOut.nResult = OTXDSP_SDK_S_OK;

    sOut.nDataSample = g_sqRxData[0];

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutGetSampleS));


  } else if (nCode == DEMO2_IOCTL_TEST_EXTERNAL_MEM) {
    Demo2IoOutTestExternalMemS sOut;
#ifdef EXTMEM_AVAILABLE
    Uint32 nDataErrors;
    Uint32 nMemErrors;

    sOut.nResult = DemoCheckExternalMemory(&nDataErrors, &nMemErrors);

    sOut.nMemErrorsLowWord = (Uint16) nMemErrors;
    sOut.nMemErrorsHighWord = (Uint16)(nMemErrors >> 16);
    sOut.nDataErrorsLowWord = (Uint16) nDataErrors;
    sOut.nDataErrorsHighWord = (Uint16)(nDataErrors >> 16);
#else // If we don't have external memory available then we can't run this IO code
    sOut.nResult = OTXDSP_SDK_E_NOT_SUPPORTED;

    sOut.nMemErrorsLowWord = 1;
    sOut.nMemErrorsHighWord = 0;
    sOut.nDataErrorsLowWord = 1;
    sOut.nDataErrorsHighWord = 0;
#endif

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutTestExternalMemS));


  } else if (nCode == DEMO2_IOCTL_LOOP_ALL_TS) {
    Demo2IoInLoopAllTsS *pInS = (Demo2IoInLoopAllTsS *)(OTXDSP_SDK_IOCTL_DATA_ADDR/2);

    g_bLoopAllTs = pInS->bLoopAllTs;
    if (!g_bLoopAllTs) {
      // Clear the transmit buffer from the previous looping data
      memset(&g_sqDmaTxBuf[0], 0, 2 * OTXDSP_SDK_FRAMES_PER_BUF*g_nTimeslotsPerFrame);
    }

    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

  } else if (nCode == DEMO2_IOCTL_GET_FULL_FRAME) {
    Demo2IoOutGetFullFrameS sOut;
    short i;

    sOut.nResult = OTXDSP_SDK_S_OK;

    for (i = 0; i < g_nTimeslotsPerFrame; i++) {
      sOut.nFrame[i] = g_sqFrameSample[i];
    }

    OtxDspCommandCompleteData(hHandle, (unsigned short *)&sOut, sizeof(Demo2IoOutGetFullFrameS));

  } else if (nCode == DEMO2_IOCTL_GET_DEBUG_DATA) {
    unsigned short i;

    // This is an example of a non-blocking call
    // We must confirm the IoControl command before we can send any other events
    // back to the host.
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_S_OK);

    // A non-blocking call would normally take some time to execute.
    // Normally the application should save the Task Reference value, perform the task,
    // then call OtxDspSetHostEvent using the save Task Reference value.
    // However, in this example we call OtxDspSetHostEvent immediately.
#if 0
    //g_debugDataS.nData0 = SEM_count(&semProcessing);
    g_debugDataS.nData1 = 0; // *((volatile ioport int *) 0x0E03); //DMA_RGET(DMAGTCR);
    g_debugDataS.nData2 = g_nDmaTxFrameCounter; 
    g_debugDataS.nData3 = g_nDmaRxTimeoutCounter;       
    g_debugDataS.nData4 = (Uint16) g_nXevtCounter; 
    g_debugDataS.nData5 = (Uint16) g_nRevtCounter; 
    g_debugDataS.nData6 = g_nDmaTxTimeoutCounter; 
    g_debugDataS.nData7 = g_nDmaRxTimeoutCounter;     
#endif

#if 1
    g_debugDataS.nData0 = g_debugData[0];
    g_debugDataS.nData1 = g_debugData[1];
    g_debugDataS.nData2 = g_debugData[2];
    g_debugDataS.nData3 = g_debugData[3];
    g_debugDataS.nData4 = g_debugData[4];
    g_debugDataS.nData5 = g_debugData[5];
    //g_debugDataS.nData6 = g_debugData[6];
    //g_debugDataS.nData7 = g_debugData[7];

    g_debugDataS.nData6 = (Uint16) g_nXevtCounter; 
    g_debugDataS.nData7 = (Uint16) g_nRevtCounter;     
#endif

    for (i = 0; i < 32; i++) {
      g_debugDataS.sqBuf[i] = g_sqDmaRxBuf[g_iTemp + i];
    }   
        
    //g_iTemp += 32;

    OtxDspSetHostEvent(DEMO2_NOTIFY_DEBUG_DATA, hHandle, nTaskRef, (unsigned short *)&g_debugDataS, sizeof(Demo2NotifyGetDebugDataS));

  } else {
    OtxDspCommandCompleteRc(hHandle, OTXDSP_SDK_E_INVALIDARG);
  }

}




/*--------------------------------------------------------------------------*\
  @func

    Call-back routine call when we have receive interrupt from the
    serial port indicating the the receive buffer is either full or
    half-full.

    This function is called from the processing task.
    TO BE RE-IMPLEMENTED BY THE USER.

\*--------------------------------------------------------------------------*/
void OtxDspServiceRxSerialPort(
  void
)
{
  short n;
  unsigned short *pRxData;

  g_nFrameCounter += OTXDSP_SDK_FRAMES_PER_BUF;

  g_nDmaRxFrameCounter++;

  if (g_bDmaRxHalf) {
    // First half (ping)
    pRxData = &g_sqDmaRxBuf[0];
  } else {               
    // Second half (pong)
    pRxData = &g_sqDmaRxBuf[OTXDSP_SDK_FRAMES_PER_BUF*g_nTimeslotsPerFrame];
  } 
  
  // Copy one full frame into a temporary storage buffer
  // Used by the DEMO2_IOCTL_GET_FULL_FRAME command
  memcpy(&g_sqFrameSample[0], pRxData, g_nTimeslotsPerFrame);

  // At this point there are OTXDSP_SDK_FRAMES_PER_BUF (25) frames of data to pick up
  g_iRxData = 0;
  for (n = 0; n < OTXDSP_SDK_FRAMES_PER_BUF; n++) {
    if (g_bRxEnabled) {
      UnpackFrame(pRxData + n*g_nTimeslotsPerFrame);  // There are g_nTimeslotsPerFrame words in one frame
    }
  } 
  
  ProcessRxSamples();
  g_bRxDataAvailable = OTXDSP_TRUE;
  
}


/*--------------------------------------------------------------------------*\
  @func

    Picks up a certain timeslot from a frame, then expands the 8-bit timeslot
    to a 16-bit value using either u-law or a-law expansion (14-bit value with
    the 2 LSB set to 0).

    This function is called from OtxDspServiceRxSerialPort().

    Note that all timeslots arrive in least significant octet of each word 
    in pFrame.

\*--------------------------------------------------------------------------*/
static void UnpackFrame(
  unsigned short *pFrame
)
{
  unsigned short nTimeSlotOctet;

  //g_debugDataS.nData0++;

  nTimeSlotOctet = *(pFrame + g_nRxTimeSlot);
  
  // Expand according to current law and store the data in our receive buffer
  g_sqRxData[g_iRxData++] = OtxDspLawExpand(nTimeSlotOctet, g_law);

}


/*--------------------------------------------------------------------------*\
  @func

    Call-back routine call when we have transmit interrupt from the
    Buffered Serial Port indicating the the transmit buffer is either empty or
    half-empty.

    This function is called from the processing task.
    TO BE RE-IMPLEMENTED BY THE USER.

\*--------------------------------------------------------------------------*/
void OtxDspServiceTxSerialPort(
  void
)
{
  short n;
  unsigned short *pTxData;

  g_nDmaTxFrameCounter++;

  if (g_bDmaTxHalf) {
    // First half (ping)
    pTxData = &g_sqDmaTxBuf[0];
  } else {               
    // Second half (pong)
    pTxData = &g_sqDmaTxBuf[OTXDSP_SDK_FRAMES_PER_BUF*g_nTimeslotsPerFrame];
  }  

  if (g_bLoopAllTs) {
    // Loop all of the received data directly to the transmit data without processing
    unsigned short *pRxData;
    if (g_bDmaRxHalf) {
      // First half (ping)
      pRxData = &g_sqDmaRxBuf[0];
    } else {               
      // Second half (pong)
      pRxData = &g_sqDmaRxBuf[OTXDSP_SDK_FRAMES_PER_BUF*g_nTimeslotsPerFrame];
    } 

    for (n = 0 ; n < (OTXDSP_SDK_FRAMES_PER_BUF*g_nTimeslotsPerFrame); n++) {
      pTxData[n] = pRxData[n];
    }

  } else {
    // Time to fill the transmit buffer with data for OTXDSP_SDK_FRAMES_PER_BUF frames
    g_iTxData = 0;
    for (n = 0; n < OTXDSP_SDK_FRAMES_PER_BUF; n++) {
      if (g_bTxEnabled) {
        PackFrame(pTxData + n*g_nTimeslotsPerFrame);
      }
    }    
    
    GenerateTxSamples();
  }

}


/*--------------------------------------------------------------------------*\
  @func

    Performs law compression on the outgoing data values. The compressed octet
    is then stored in the corresponding timeslot in the frame.

    This function is called from OtxDspServiceRxSerialPort().
    
\*--------------------------------------------------------------------------*/
static void PackFrame(
  unsigned short *pFrame
)
{
  if (g_law == OTXDSP_ALAW) {
    // A-law
    *(pFrame + g_nTxTimeSlot) = OtxDspAlawCompress(g_sqTxData[g_iTxData++]);
  } else {
    // U-law  
    *(pFrame + g_nTxTimeSlot) = OtxDspUlawCompress(g_sqTxData[g_iTxData++]);    
  }
}



/*--------------------------------------------------------------------------*\
  @func

    Process Received Samples

\*--------------------------------------------------------------------------*/
void ProcessRxSamples(
  void
)
{
  if (g_bStreamRxData) {
    // Send the received data stream to the host
    static unsigned short nTaskRef = 0;
    Demo2NotifyStreamRxDataS streamRxDataS;

    memcpy(streamRxDataS.nData, g_sqRxData, 25);
    OtxDspSetHostEvent(DEMO2_NOTIFY_STREAM_RX_DATA, g_hHandle, nTaskRef++, (unsigned short *)&streamRxDataS, sizeof(Demo2NotifyStreamRxDataS));
  }
}


/*--------------------------------------------------------------------------*\
  @func

    Generate more samples to Transmit

\*--------------------------------------------------------------------------*/
void GenerateTxSamples(
  void
)
{
  short i;

  // Simply copy the received samples to the outgoing transmit data buffer
  // Perform a gain adjustment for each copied sample
  
  if (!g_bRxDataAvailable) {
    g_debugDataS.nData0++;
  }

  for (i = 0; i < OTXDSP_SDK_FRAMES_PER_BUF; i++) {
    g_sqTxData[i] = g_sqRxData[i] << g_nTxGain;
  }
  
  g_bRxDataAvailable = OTXDSP_FALSE;

}


#ifdef EXTMEM_AVAILABLE
/*--------------------------------------------------------------------------*\
  @func

    Check external memory by writing and reading to it.
    
    The external memory starts at byte adress 0x50000 and contains
    0xFA8000 bytes.
    
    This function hogs the CPU and should really be executing as a separate
    task with a lower priority.
    
\*--------------------------------------------------------------------------*/
static unsigned short DemoCheckExternalMemory(
  Uint32 *nDataErrors,
  Uint32 *nMemErrors
)
{
  unsigned short dspRc = OTXDSP_SDK_S_OK; // Assume successful test
	Uint32 nDataErr = 0;
	Uint32 nMemErr = 0;
	Uint16 i;     
	Uint32 nMemIndex;     
	Uint32 *pMem; 
	Uint32 nMemExtLen = 0xFA8000; // in bytes
	Uint32 nWriteVal;
  
  // Test all data bits
  for (i = 0; i < 16; i++) {
    g_sqBigBuf[i] = (Uint16) 0x1 << i ;
  }    
  
  for (i = 0; i < 16; i++) {
    if (g_sqBigBuf[i] != ((Uint16) 0x1 << i)) {
      nDataErr++;
    }  
  }
   
  LOG_printf(&trace, "g_sqBigBuf[10] = %d. nDataErr=%ld\n", g_sqBigBuf[i], nDataErr);
  
  // Test all addresses of the extended memory
  for (nMemIndex = 0; nMemIndex < nMemExtLen; nMemIndex += 4) {
    nWriteVal = (Uint32) nMemIndex + 0x50000;
    pMem = (Uint32 *)(nWriteVal/2);
    *pMem = nWriteVal;
  }

  for (nMemIndex = 0; nMemIndex < nMemExtLen; nMemIndex += 4) {
    nWriteVal = (Uint32) nMemIndex + 0x50000;
    pMem = (Uint32 *)(nWriteVal/2);
    
    if (*pMem != nWriteVal) {
      nMemErr++;
    }  
  }      
  
  LOG_printf(&trace, "Mem nMemErr=%d\n", nMemErr);
  
  if (nDataErr || nMemErr) {
    dspRc = OTXDSP_SDK_S_FALSE; // Errors occurred 
  } 
  
  *nDataErrors = nDataErr;
  *nMemErrors  = nMemErr;
  
  return dspRc;

}
#endif
